/* Nessus
 * Copyright (C) 1998 - 2003 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * In addition, as a special exception, Renaud Deraison
 * gives permission to link the code of this program with any
 * version of the OpenSSL library which is distributed under a
 * license identical to that listed in the included COPYING.OpenSSL
 * file, and distribute linked combinations including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * this file, you may extend this exception to your version of the
 * file, but you are not obligated to do so.  If you do not wish to
 * do so, delete this exception statement from your version.
 *
 * Attack.c : 
 *  launch the plugins, and manages the multithreading 
 *
 */
 
#include <includes.h>
#include "attack.h"
#include "log.h"
#include "hosts_gatherer.h"
#include "sighand.h"
#include "rules.h"
#include "auth.h"
#include "threads.h"
#include "comm.h" 
#include "utils.h"
#include "preferences.h"
#include "ntp.h"
#include "ntp_11.h"
#include "pluginload.h"
#include "pluginlaunch.h"
#include "plugs_req.h"
#include "save_tests.h"
#include "save_kb.h"
#include "detached.h"
#include "pluginscheduler.h"

#define ERR_HOST_DEAD -1
#define ERR_CANT_FORK -2

#define MAX_FORK_RETRIES 10

extern u_short * getpts(char *);

/*******************************************************

		PRIVATE FUNCTIONS
		
********************************************************/


static struct nessusd_threads ** Threads  = NULL;
static void
attack_sigterm()
{
 if(Threads)
  nessusd_thread_kill_all(Threads);

 wait_for_children();
}


static void 
arg_addset_value(arg, name, type, len, value)
 struct arglist *arg;
 char * name;
 int type, len;
 void * value;
{
 if(arg_get_type(arg, name) < 0)
  arg_add_value(arg, name, type, len, value);
 else
  arg_set_value(arg, name, len, value);
}



/*-------------------------------------------------------

	Init. an arglist which can be used by 
	the plugins from an hostname and its ip
	
--------------------------------------------------------*/	
static struct arglist * 
attack_init_hostinfos(mac, hostname, ip)
     char * mac;
     char * hostname;
    struct in_addr * ip;
{
  struct arglist * hostinfos;
  struct arglist * ports;
  struct in_addr addr;
  
  hostinfos = emalloc(sizeof(struct arglist));
  if(inet_aton(hostname, &addr) != 0)
  {
   char * f = (char*)hg_get_name_from_ip(addr);
   arg_add_value(hostinfos, "FQDN", ARG_STRING, strlen(f), estrdup(f));
   free(f);
  }
  else
   arg_add_value(hostinfos, "FQDN", ARG_STRING, strlen(hostname), hostname);
   
   
  if(mac)
  {
  	arg_add_value(hostinfos, "NAME", ARG_STRING, strlen(mac), mac);
	arg_add_value(hostinfos, "MAC", ARG_STRING, strlen(mac), mac);
  }
  else
  	arg_add_value(hostinfos, "NAME", ARG_STRING, strlen(hostname), hostname);
	
  arg_add_value(hostinfos, "IP", ARG_PTR, sizeof(struct in_addr), ip);
  ports = emalloc(sizeof(struct arglist));
  arg_add_value(hostinfos, "PORTS", ARG_ARGLIST, sizeof(struct arglist), ports);
  return(hostinfos);
}

/*--------------------------------------------------------
	
		 Return our user name
 
 ---------------------------------------------------------*/
 
static char *
attack_user_name(globals)
 struct arglist * globals;
{
 static char * user;
 if(!user)
   user = (char*)arg_get_value(globals, "user");
  
 return user;
}


static int
launch_plugin(globals, plugins, hostname, cur_plug, num_plugs, key, hostinfos, portsinfos, new_kb)
 struct arglist * globals;
 struct arglist * plugins;
 char * hostname;
 int *cur_plug, num_plugs;
 struct arglist * key;
 struct arglist * hostinfos;
 struct arglist * portsinfos;
 int new_kb;
{
  struct arglist * preferences = arg_get_value(globals,"preferences");
  struct arglist * args = plugins->value;
  char * name;
  int optimize = preferences_optimize_test(preferences);
  int category = plug_get_category(args);
      
  name = estrdup(plug_get_path(args));
  if(plug_get_launch(args) || 
     category == ACT_INIT ||
    (category == ACT_SETTINGS)) /* can we launch it ? */
  {
   char * error;
   
   pl_class_t * cl_ptr = arg_get_value(plugins->value, "PLUGIN_CLASS");
   
  
  if(preferences_safe_checks_enabled(preferences) && 
  	(category == ACT_DESTRUCTIVE_ATTACK ||
	 category == ACT_KILL_HOST ||
	 category == ACT_DENIAL))
	 	{
		if(preferences_log_whole_attack(preferences))
		  log_write("user %s : Not launching %s against %s %s (this is not an error)\n",
	       			attack_user_name(globals),
				plugins->name, 
				hostname, 
				"because safe checks are enabled");
		plugin_set_running_state(plugins->value, PLUGIN_STATUS_DONE);
		efree(&name);		
		return 0;
		}
		
   comm_send_status(globals, hostname, "attack", (*cur_plug)++, num_plugs);


    if(save_kb(globals))
    {
     int id = plug_get_id(plugins->value);
     char asc_id[30];
	 
     
     /*
      * Does our user want a differential scan only ?
      */
      if(diff_scan(globals))
        diff_scan_enable(plugins->value);
	 
	 
	 			     
	 sprintf(asc_id, "Launched/%d", id);
	 if(arg_get_value(key, asc_id) &&
	    !save_kb_replay_check(globals, category))
	  {
	   /* 
	    * XXX determine here if we should skip
	    * ACT_SCANNER, ACT_GATHER_INFO, ACT_ATTACK and ACT_DENIAL
	    */
	   if(preferences_log_whole_attack(preferences))
	    log_write("user %s : Not launching %s against %s %s (this is not an error)\n",
	       			attack_user_name(globals),
				plugins->name, 
				hostname, 
				"because it has already been launched in the past");
	   plugin_set_running_state(plugins->value, PLUGIN_STATUS_DONE);			
	   return 0;
	  }
	  else {
	  	arg_addset_value(key, asc_id, ARG_INT, sizeof(int), (void*)1);
		save_kb_write_int(globals, hostname, asc_id,  1);
		}
       }	     
	
	 
	 if(!optimize || 
	   !(error = requirements_plugin(key, plugins,portsinfos, preferences)))
	 {
	  int pid;
	  
	 /*
	  * Start the plugin
	  */
	 pid = plugin_launch(globals, plugins, hostinfos, preferences, key, name, cl_ptr);
	 if(pid  < 0)	
	 	{
		plugin_set_running_state(plugins->value, PLUGIN_STATUS_UNRUN);
		return ERR_CANT_FORK;
		}
		 
	 if(preferences_log_whole_attack(preferences))
	 	log_write("user %s : launching %s against %s [%d]\n", 
	 				attack_user_name(globals),
					plugins->name, 
					hostname,
					pid);
					
	 	     
	 /*
	  * Stop the test if the host is 'dead'
	  */	 
        if(arg_get_value(key, "Host/dead") ||
	   arg_get_value(key, "Host/ping_failed"))
	{
	  log_write("user %s : The remote host (%s) is dead\n",
	  				attack_user_name(globals),
	  				hostname);
	  pluginlaunch_stop();		
	  if(new_kb)save_kb_close(globals, hostname);	
	  if(arg_get_value(key, "Host/ping_failed"))
	  {
	   save_kb_restore_backup(globals, hostname);
	  }
	  
	  efree(&name);
	  plugin_set_running_state(plugins->value, PLUGIN_STATUS_DONE);	/* XXX ? */				
	  return ERR_HOST_DEAD;
	}
       }
       
       else /* requirements_plugin() failed */
	  {
	   plugin_set_running_state(plugins->value, PLUGIN_STATUS_DONE);
	   if(preferences_log_whole_attack(preferences))
	    log_write("user %s : Not launching %s against %s %s (this is not an error)\n",
	       			attack_user_name(globals),
				plugins->name, 
				hostname, 
				error);
	
	  }
      } /* if(plugins->launch) */
      else    
       plugin_set_running_state(plugins->value, PLUGIN_STATUS_DONE);
       
      efree(&name); 
      return 0;
}

/*--------------------------------------------------------
	
	          Attack _one_ host

----------------------------------------------------------*/	
static void 
attack_host      (globals, 
		  hostinfos, 
		  plugins, 
		  hostname, 
		  port_range, 
		  hosts_list,
		  sched)
     
     
     struct arglist * globals;
     struct arglist * hostinfos;
     struct arglist * plugins;
     char * hostname;
     char * port_range;
     struct arglist * hosts_list;
     plugins_scheduler_t sched;
{ 

  /*
   * Used for the status
   */
  int num_plugs = 0;
  int cur_plug = 1;
  
  struct arglist * key;
  struct arglist * portsinfos = emalloc(sizeof(struct arglist));
  int new_kb = 0;
  int kb_restored = 0;
 int forks_retry = 0;
 
 
  setproctitle("testing %s", (char*)arg_get_value(hostinfos, "NAME"));
  comm_send_status(globals, hostname, "attack", 0, 100);
  
  if(save_kb(globals))
  {
   if(
      save_kb_exists(globals, hostname) &&
      save_kb_pref_restore(globals)
      )
   {
    save_kb_backup(globals, hostname);
    key = save_kb_load_kb(globals, hostname);
    /*
     * XXXX
     * - Mark the KB as restored
     *
     * - Should send a warning telling that no port scan has
     *   been fully performed this time
     */
    kb_restored = 1; 
  }
  else 
  {
   save_kb_new(globals, hostname);
   key = emalloc(sizeof(struct arglist));
   new_kb = 1;
   }
  
  /* XXX */
  arg_add_value(globals, "CURRENTLY_TESTED_HOST", ARG_STRING, strlen(hostname), hostname);
 }
 else key = emalloc(sizeof(struct arglist));
  
  
  arg_add_value(key, "hosts", ARG_ARGLIST, -1, hosts_list);
  
  

  num_plugs = get_active_plugins_number(plugins);

  
 
  arg_add_value(portsinfos, "ports", ARG_PTR, -1, arg_get_value(globals,"ports"));
  arg_add_value(portsinfos, "ports_num", ARG_INT, -1, arg_get_value(globals,
  								"ports_num"));
  arg_add_value(portsinfos, "key", ARG_ARGLIST, -1, key);
  
  
  
  /* launch the plugins */
  pluginlaunch_init(globals);
 
   
   for(;;)
   {
    struct arglist * plugin;
    plugin = plugins_scheduler_next(sched);
    if(plugin != NULL && plugin != PLUG_RUNNING)
    { 
      int e;
 again:        
      if((e = launch_plugin(globals, 
      		    plugin, 
		    hostname, 
		    &cur_plug, 
		    num_plugs, 
		    key, 
		    hostinfos, 
		    portsinfos, 
		    new_kb))  < 0)
		    {
		     /*
		      * Remote host died
		      */
		     if(e == ERR_HOST_DEAD)
		     	goto host_died;
		     else if(e == ERR_CANT_FORK )
		     {
		      if(forks_retry < MAX_FORK_RETRIES)
		      {
		       log_write("fork() failed - waiting a little");
		       sleep(5);
		       forks_retry++;
		       goto again;
		      }
		      else {
		      	log_write("fork() failed too many times - aborting");
			goto host_died;
			}
		     }
		    }
      }
     else if(plugin == NULL) break;
     else pluginlaunch_wait_for_free_process();
    }
  pluginlaunch_wait();
host_died:  
  pluginlaunch_stop();
  plugins_scheduler_free(sched);
  if(new_kb)save_kb_close(globals, hostname);
  efree(&portsinfos);
}

/*-----------------------------------------------------------------

  attack_start : set up some data and jump into
  attack_host()

 -----------------------------------------------------------------*/
static void
attack_start(args)
  struct arglist * args;
{
 struct arglist * globals = arg_get_value(args, "globals");
 char * hostname = arg_get_value(args, "hostname");
 char * mac = arg_get_value(args, "host_mac_addr");
 struct arglist * plugs = arg_get_value(args, "plugins");
 char * port_range = arg_get_value(args, "port_range");
 struct in_addr * hostip = arg_get_value(args, "hostip");
 struct arglist * hosts_list = arg_get_value(args, "hosts_list");
 struct arglist * hostinfos;
 
 struct arglist * preferences = arg_get_value(globals,"preferences");
 char * non_simult = arg_get_value(preferences, "non_simult_ports");
 int thread_socket = (int)arg_get_value(globals, "thread_socket");
 int soc;
 struct arglist * new_plugs = NULL;
 struct attack_atom ** atoms = arg_get_value(args, "atoms");
 struct timeval then, now;
 plugins_scheduler_t * sched = (plugins_scheduler_t)arg_get_value(args, "sched");
 gettimeofday(&then, NULL);
 attack_atom_free_others(atoms, hostname);

 if(!non_simult){
        non_simult = estrdup("139, 445");
 	arg_add_value(preferences, "non_simult_ports", ARG_STRING, strlen(non_simult), non_simult);
	}
 arg_add_value(preferences, "non_simult_ports_list", ARG_ARGLIST, -1, (void*)list2arglist(non_simult));
 /*
  * Options regarding the communication with our father
  */
 set_linger(thread_socket, 60);
 nessus_deregister_connection((int)arg_get_value(globals, "global_socket"));
 arg_set_value(globals, "global_socket", -1, (void*)thread_socket);
 
 /*
  * Wait for the server to confirm it read our data
  * (prevents client desynch)
  */
 arg_add_value(globals, "confirm", ARG_INT, sizeof(int), (void*)1);
 
 soc = thread_socket;
 hostinfos = attack_init_hostinfos(mac, hostname,hostip);
 if(mac)
  hostname = mac;
  
#ifndef USE_FORK_THREADS
   new_plugs = emalloc(sizeof(struct arglist));
   arg_dup(new_plugs, plugs);
#else
    new_plugs = plugs;
#endif
  plugins_set_socket(plugs, soc);
  ntp_1x_timestamp_host_scan_starts(globals, hostname);
  attack_host(globals, hostinfos, new_plugs, hostname, port_range, hosts_list, sched);
  if(preferences_ntp_show_end(preferences))ntp_11_show_end(globals, hostname);
  ntp_1x_timestamp_host_scan_ends(globals, hostname);
  gettimeofday(&now, NULL);
  if(now.tv_usec < then.tv_usec)
  {
   then.tv_sec ++;
   now.tv_usec += 1000000;
  }
  log_write("Finished testing %s. Time : %ld.%.2ld secs\n",
  		hostname,
  		now.tv_sec - then.tv_sec,
		(now.tv_usec - then.tv_usec) / 10000);
  shutdown(soc, 2);
  close(soc);
  
}

/*******************************************************

		PUBLIC FUNCTIONS
		
********************************************************/


/*------------------------------------------------

   This function attacks a whole network
 		
 -----------------------------------------------*/
int 
attack_network(globals)
    struct arglist * globals;
{
  int max_hosts			= 0;
  int num_tested		= 0;
  int host_pending		= 0;
  int num_threads		= 0;
  char * hostname 		= NULL;
  struct in_addr host_ip	= {0};
  char * port_range		= NULL;
  int hg_flags			= 0;
  struct hg_globals * hg_globals = NULL;
  int global_socket		= -1;
  struct arglist * preferences  = NULL;
  struct arglist * plugins      = NULL;
  struct arglist * hosts_list   = NULL;
  struct attack_atom ** atoms   = NULL;
  ntp_caps* caps		= NULL;
  struct nessusd_threads ** threads = NULL;
  struct nessus_rules *rules	= NULL;
  struct arglist * rejected_hosts =  NULL;
  unsigned short * ports	= NULL;
  int data			= 0;
  int restoring    = 0;
  harglst * tested = NULL;
  int  save_session= 0;  
  int continuous   = 0;
  int detached     = 0;
  hargwalk *  hw;
  harglst * files;
  char * key;
  int return_code = 0;
  plugins_scheduler_t sched;

  
  preferences    = arg_get_value(globals, "preferences");
  detached = preferences_detached_scan(preferences);
  if(detached)
  {
   char *email;
   
   detached_new_session(globals, arg_get_value(preferences, "TARGET"));
   arg_addset_value(preferences, "ntp_keep_communication_alive", ARG_STRING, 0, NULL);
   
   /* 
    * Tell the client that the scan is finished (actually, we will
    * work hard to test the network, but this lazy client has the right
    * to get some rest...
    */
   log_write("user %s : running a detached scan\n", (char*)arg_get_value(globals, "user"));
   global_socket = (int)arg_get_value(globals, "global_socket");
   comm_terminate(globals);
   close_stream_connection(global_socket);
   global_socket = -1;
   arg_set_value(globals, "global_socket", sizeof(int), (void*)global_socket);
   
   nessus_signal(SIGTERM, attack_sigterm);
   
start_attack_network:
   if((email = preferences_detached_scan_email(preferences)))
   {
    /*
     * The user wants to receive the results by e-mail.
     */
    if(detached_setup_mail_file(globals, email))
    {
     /*
      * Could not create the file to store our data
      */
     close_stream_connection(global_socket);
     return -1;
    }
   }
   else arg_addset_value(globals, "detached_scan_email_address", ARG_STRING, 0, NULL);
  }  
  num_tested = 0;
  num_threads = 0;
  
  global_socket  = (int)arg_get_value(globals, "global_socket");
 
  plugins        = arg_get_value(globals, "plugins");
  hosts_list     = emalloc(sizeof(struct arglist));
  atoms          = attack_atom_new();
  caps           = arg_get_value(globals, "ntp_caps");
  rules          = arg_get_value(globals, "rules");
  rejected_hosts = emalloc(sizeof(struct arglist));
  
  if(detached)
    continuous = preferences_continuous_scan(preferences);
  
  save_session = preferences_save_session(preferences);
  if(continuous)
   restoring = 0;
  else
   restoring = ((int)arg_get_value(globals, "RESTORE-SESSION") == 1);
   
  if(restoring)tested = arg_get_value(globals, "TESTED_HOSTS");
  if(save_session)save_tests_init(globals);  


  hostname = arg_get_value(preferences, "TARGET");
  if(!hostname){
  	log_write("%s : TARGET not set ?!", 
			attack_user_name(globals));
	EXIT(1);
	}		
  
  if(port_range)efree(&port_range);
  
  port_range = arg_get_value(preferences, "port_range");
  if(!port_range||!strlen(port_range))port_range = estrdup("1-15000");
  else port_range = estrdup(port_range);
  
  if(strcmp(port_range, "-1"))
  {
   int num = 0;
   int i = 0;
   if(ports)efree(&ports);
   ports = (unsigned short*)getpts(port_range);
   if(!ports){
   	auth_printf(globals, "SERVER <|> ERROR <|> Invalid port range <|> SERVER\n");
	return -1; 
	}
   for(i=0;ports[i];i++,num++);
   if(arg_get_value(globals, "ports"))
    arg_set_value(globals, "ports", -1, ports);
   else
    arg_add_value(globals, "ports", ARG_PTR,-1, ports);
    
   if(arg_get_value(globals, "ports_num"))
    arg_set_value(globals, "ports_num", sizeof(int),  (void*)num);
   else
    arg_add_value(globals, "ports_num", ARG_INT, sizeof(int), (void*)num);
  }
  /*
   * Initialization of the attack
   */
  sched  = plugins_scheduler_init(plugins,  preferences_autoload_dependencies(preferences) ); 
  if(caps->ntp_11)
    send_plugin_order(globals, plugins);
  
  
  
  hg_flags = preferences_get_host_expansion(preferences);

  max_hosts = get_max_hosts_number(preferences);
  if(!restoring)
  {
  log_write("user %s starts a new attack. Target(s) : %s, with max_hosts = %d\n",
			 attack_user_name(globals), 
			 hostname,
			 max_hosts);

  }
  else
  {
   log_write("user %s restores session %s, with max_hosts = %d\n",
   			attack_user_name(globals),
			(char*)arg_get_value(globals, "RESTORE-SESSION-KEY"),
			max_hosts);
			
   save_tests_playback(globals, arg_get_value(globals, "RESTORE-SESSION-KEY"),tested);
  }
  
  			 
  /* 
   * Initialize the hosts_gatherer library 
   */
   			  
  hg_globals = hg_init(hostname, hg_flags);
  hostname = hg_next_host(hg_globals,&host_ip);
  if(tested)
   while(hostname && 
         harg_get_int(tested, hostname)){
	 		efree(&hostname);
			hostname = hg_next_host(hg_globals, &host_ip);
			}
  if(hostname)
  {			 
   Threads = threads = nessusd_threads_init();
   if(arg_get_value(globals, "threads"))
    arg_set_value(globals, "threads", -1, threads);
   else
    arg_add_value(globals, "threads", ARG_PTR, -1, threads);
  
 
  
  /*
   * Start the attack !
   */
   
#if 0
   if(arg_get_value(globals, "client_socket"))
    arg_set_value(globals, "client_socket", sizeof(int), (void*)global_socket);
   else
    arg_add_value(globals, "client_socket", ARG_INT, sizeof(int), (void*)global_socket);       
#endif   
   if(arg_get_value(globals, "father"))
    arg_set_value(globals, "father", sizeof(int), (void*)getpid());
   else
    arg_add_value(globals, "father", ARG_INT, sizeof(int), (void*)getpid());
   
   
   while(hostname)
    {
      nthread_t pid;
      
      
   /*
    * nessusd offers the ability to either test
    * only the hosts we tested in the past, or only
    * the hosts we never tested (or both, of course)
    */
   if(save_kb(globals))
    {
    if(save_kb_pref_tested_hosts_only(globals))
    {
    if(!save_kb_exists(globals, hostname))
     {
      log_write("user %s : not testing %s because it has never been tested before\n",
      		 (char*)arg_get_value(globals, "user"), 
		 hostname);
      efree(&hostname);
      hostname = hg_next_host(hg_globals, &host_ip);
      continue;
     }
   }
   else if(save_kb_pref_untested_hosts_only(globals))
   {
    /* XXX */
    if(save_kb_exists(globals, hostname))
    {
     log_write("user %s : not testing %s because it has already been tested before\n", 
     			(char*)arg_get_value(globals, "user"), 
			hostname);
     efree(&hostname);			
     hostname = hg_next_host(hg_globals, &host_ip);
     continue;
    }
   }
  }

      host_pending = 0 ;
      if(CAN_TEST(get_host_rules(rules, host_ip,32))) /* do we have the right to test this host ? */ 
      {
      
	  if(num_threads < max_hosts)
	    {
             struct arglist * arglist = emalloc(sizeof(struct arglist));
	     struct in_addr * ia = emalloc(sizeof(ia));
	     int s[2] = {-1,-1};
	     char * MAC = NULL;
	     int mac_err = -1;
	     
	     
	    if(preferences_use_mac_addr(preferences) &&
	        is_local_ip(host_ip))
	     {
	      mac_err = get_mac_addr(host_ip, &MAC);
	      if(mac_err > 0)
	      {
	       /* remote host is down */
	       hostname = hg_next_host(hg_globals, &host_ip);
	       continue;
	      }
	     }
	     
	     if(!socketpair(AF_UNIX, SOCK_STREAM, 0, s))
	     {
	      if(arg_get_value(globals, "thread_socket"))
	       arg_set_value(globals, "thread_socket", sizeof(int), (void*)s[0]);
	      else
	       arg_add_value(globals, "thread_socket", ARG_INT, sizeof(int), (void*)s[0]);
	      }
	     else perror("socketpair ");	
	     arg_add_value(hosts_list, hostname, ARG_INT, -1, (void *)1);
             arg_add_value(arglist, "globals", ARG_ARGLIST, -1, globals);
             arg_add_value(arglist, "hostname", ARG_STRING, -1, hostname);
	     ia->s_addr = host_ip.s_addr;
	     arg_add_value(arglist, "hostip", ARG_PTR, -1, ia); 
	     if(MAC) arg_add_value(arglist, "host_mac_addr", ARG_STRING, -1, MAC);
	     
             arg_add_value(arglist, "plugins", ARG_ARGLIST, -1, plugins);
             arg_add_value(arglist, "port_range", ARG_STRING, -1, port_range);
             arg_add_value(arglist, "hosts_list", ARG_ARGLIST, -1, hosts_list); 
	     arg_add_value(arglist, "atoms", ARG_PTR, -1, atoms);
	     arg_add_value(arglist, "sched",  ARG_PTR, -1, sched);
	     pid = create_thread((void*)attack_start, arglist, (struct arglist*)-1); 
	     if(pid < 0)
	      {
	      /*
	       * forking failed - we go to the wait queue
	       */
	      efree(&MAC);
	      efree(&ia);
	      arg_free(arglist);
	      goto wait_queue;
	      }
	     nessusd_thread_register(threads, MAC ? MAC : hostname, pid);
	     if(s[0] >= 0)
	  	   attack_atom_insert(atoms, MAC ? MAC : hostname, s[1], s[0]);
             num_threads++;
	     log_write("user %s : testing %s [%d]\n", attack_user_name(globals), hostname, pid);
	     arg_free(arglist);
	     efree(&MAC);
	     efree(&ia);
	    }
	  else 
	    {
wait_queue:	    
              /* 
               * There are too many threads at this time... We
               * have to wait until one of them finishes
               */
	      host_pending = 1; /* means that we must NOT update the hostname */
             
	      while(num_threads >= max_hosts)
                {
		 int e;
		/*
		 * Check the threads input
		 */
		if((e = check_threads_input(atoms, detached ? -1 : global_socket, globals))<0)
			{
			if(e == -1)
			{
			 nessusd_thread_kill_all(threads);
			 return_code = -1;
			}
			goto stop;		
			}
		num_threads = wait_for_a_thread(threads);
                }
	    }
	} else {
		log_write("user %s : rejected attempt to scan %s", 
			attack_user_name(globals), hostname);
		  arg_add_value(rejected_hosts, hostname, ARG_INT, sizeof(int), (void*)1);	
		}	
      if(!host_pending)
      {
        num_tested++;
	efree(&hostname);
       	hostname = hg_next_host(hg_globals, &host_ip);
	if(tested)
	 while(hostname &&
	       harg_get_int(tested, hostname))
		{
		efree(&hostname);
		hostname = hg_next_host(hg_globals, &host_ip);
		}
	
      }
     }
 
   
    
 
  /*
   * Every host is being tested... We have to wait for the threads
   * to terminate
   */
  
  
  while(wait_for_a_thread(threads)){
	/*
	 * check threads input
	 */
	int n = check_threads_input(atoms, detached ? -1 : global_socket, globals);
	while(n)
	 {
	 n = check_threads_input(atoms, detached ? -1 : global_socket, globals);
	 if(n < 0)
	  {
	  if(n == -1)
	  {
	  nessusd_thread_kill_all(threads);
	  return_code = -1;
	  }
	  goto stop;
	  }
	 wait_for_a_thread(threads);
	 }  
        }
  
   data = check_threads_input(atoms, detached ? -1 : global_socket, globals);
   if(data < 0)
    {
    if(data == -1)
    {
    nessusd_thread_kill_all(threads);
    return_code = -1;
    }
    goto stop;
    }
    
   while(data>0)
    {
    data = check_threads_input(atoms, detached ? -1 : global_socket, globals);
    if(data < 0)
     {
     if(data == -1)
     {
     nessusd_thread_kill_all(threads);
     return_code = -1;
     }
     goto stop;
     }
    } 
      
    
    
    
   /*
    * some_function(hosts_arglist_to_string(tested_hosts));
    */
    
   log_write("user %s : test complete", attack_user_name(globals));
   
    /*
     * Delete the files uploaded by the user, if any
     */
    files = arg_get_value(globals, "files_translation");
    if(files)
    {
     hw  = harg_walk_init(files);
     while((key = (char*)harg_walk_next(hw)))
      {
      unlink(harg_get_string(files, key));
      }
    }
  }
 
  if(rejected_hosts && rejected_hosts->next)
   {
     char * banner = emalloc(4001);
     int length = 0;

     sprintf(banner, "SERVER <|> ERROR <|> These hosts could not be tested because you are not allowed to do so :;");
     length = strlen(banner);

     while(rejected_hosts->next && (length < (4000-3)))
	{
	  int n;
	  n = strlen(rejected_hosts->name);
	  if(length + n + 1 >= 4000)
	  {
	    n = 4000 - length  - 2;
	  }
	  strncat(banner, rejected_hosts->name, n);
	  strncat(banner, ";", 1);
	  length+=n+1;
	  rejected_hosts = rejected_hosts->next;
	}
      if(rejected_hosts->next)
       strcat(banner, "...");
       
     auth_printf(globals, "%s\n", banner);
   }
	
stop:
  if(save_session){
  	save_tests_close(globals);
	if(!preferences_save_empty_sessions(preferences))
	{
	 if(save_tests_empty(globals))
          {
            log_write("user %s : Nothing interesting found - deleting the session\n",
    		(char*)arg_get_value(globals, "user"));
             save_tests_delete_current(globals);
	  }
        }
   }
  
  hg_cleanup(hg_globals);
  ports = arg_get_value(globals, "ports");
  efree(&ports);
  efree(&port_range);
  
  arg_free_all(hosts_list);
  
  nessusd_threads_free(threads);
  Threads = NULL;
  attack_atom_free(atoms);
  arg_free_all(rejected_hosts);
  plugins_scheduler_free(sched);
  if(detached)
  {
   if(preferences_detached_scan_email(preferences))
    detached_send_email(globals);
  }
  
  if(continuous){
  	sleep(preferences_delay_between_scans(preferences));
	plugins = plugins_reload(preferences, plugins);
	arg_set_value(globals, "plugins", -1, plugins);
  	goto start_attack_network;
	}
  else if(detached)detached_end_session(globals);	
  
  return return_code;
}

   
 
